/*
# Copyright (C) 2010
#
# This program is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
#
*/
#define ISOMODULE	"SystemControl"	// Read ISO filename from this module
#define ISOOFFSET	0x0000B108      // Offset of ISOMODULE textaddr that contains ISO filename
#define UMDIDADDR	0x880137B0	// Raw memory address that contains UMDID ( in 5.0 m33 at least )
#define CFG_MAX_ENTRIES		64      // Do you really need more than 64 entries?
#define DEFAULT_LOGTOFILE       0       // 0 = none, 1 = logging, 2 = lots-o-logging


#include <pspkernel.h>
#include <pspsdk.h>
#include <psputilsforkernel.h>
#include <pspsysmem_kernel.h>

#include <systemctrl.h>
#include <pspinit.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <psploadcore.h>
#include <pspumd.h>

#include <sys/unistd.h>

#include <psprtc.h>

PSP_MODULE_INFO("Pergame", PSP_MODULE_KERNEL, 1, 5);

//////////////////////////////////
///// wheee %data

typedef struct Cfg {
	char modname[64];
	char bootname[64];
	int level;
} Cfg;

typedef struct Loaded {
  char modname[64];
} Loaded;
//////////////////////////////////
///// whee our $variables
static unsigned int logtofile = DEFAULT_LOGTOFILE;
static unsigned int cfg_count = 0;
static Cfg * cfg = NULL;
static unsigned int loaded_count = 0;
static Loaded * loaded = NULL;

static int sema = -1;
static int cfg_mem_id = -1;

static char * eboot;
static int bootdev, apitype, keyconf;

char lstr[192];
char log_name[256];
char conf_name[256];

SceInt64 log_res = -1;
int logfd = -1;

STMOD_HANDLER previous = NULL;
int OnModuleStart(SceModule2* mod);
//////////////////////////////////
///// macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros macros
#define peekbyte(addr) (*((unsigned char*)(long)(addr)));
#define isspace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) =='\r' )
#define log(...)\
  sprintf( lstr, __VA_ARGS__ );\
  logfunc(lstr);
#define log2(...)\
  if (logtofile >= 2) {\
    sprintf( lstr, __VA_ARGS__ );\
    logfunc(lstr);\
  }
#define flog(...)\
  sprintf( lstr, __VA_ARGS__ );\
  filelog(lstr);

//////////////////////////////////
///// Append line to logfile. Still somewhat gimpy.
int filelog( const char * str ) {
  if (logfd < 0) {
    //sceIoWaitAsync(logfd,  &log_res);
    logfd = sceIoOpen( log_name, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_APPEND, 0777 );
    if (logfd >= 0) {
      //sceIoWaitAsync(logfd,  &log_res);
      sceIoWriteAsync(logfd, str, strlen( str ) );
      sceIoWaitAsync(logfd, &log_res);
      sceIoCloseAsync(logfd);
      logfd = -1;
    }
  } else {
    sceIoWaitAsync(logfd,  &log_res);
    sceIoWriteAsync(logfd, str, strlen( str ) );
  }
  return 0;
}


///////////////////////////////
///// Prepare line for logging
void logfunc( const char * str ) {
  if (logtofile > 0) {
    char lpfxstr[256];
    pspTime ctime;
    sceRtcGetCurrentClockLocalTime(&ctime);
    SceKernelThreadInfo thinfo;
    SceUID thid = sceKernelGetThreadId();
    memset(&thinfo, 0, sizeof(thinfo));                
    thinfo.size = sizeof(thinfo);      
    sceKernelReferThreadStatus(thid, &thinfo);
    int free=sceKernelPartitionTotalFreeMemSize(PSP_MEMORY_PARTITION_KERNEL);
    sprintf( lpfxstr, "%02d:%02d:%02d.%06d (%d free) - %s - %s",ctime.hour,ctime.minutes,ctime.seconds,ctime.microseconds, free ,thinfo.name, str);
    filelog(lpfxstr);
  }
}

//////////////////////////////
///// Thanks, cxmb!
int readLine( int fd, char * buf, int max_len )
{
  int i = 0, bytes;
  while( i < max_len && ( bytes = sceIoRead( fd, buf + i, 1 ) ) == 1 ) {
    if ( buf[i] == -1 || buf[i] == '\n' || buf[i] == '\r' ) break;
    buf[i]=tolower(buf[i]);
    i ++;
  }
  buf[i] = 0;
  if ( bytes != 1 && i == 0 ) return -1;
  return i;
}


//////////////////////////////////
///// Thanks, Linux!
char *strstrip(char *s) {
  size_t size;
  char *end;
  size = strlen(s);
  if (!size)
     return s;
  end = s + size - 1;
  while (end >= s && isspace(*end))
     end--;
  *(end + 1) = '\0';
  while (*s && isspace(*s))
     s++;
  return s;
}
//////////////////////////////////
///// Stuff contents of pergame.txt into cfg struct.
void parse_conf ( void ) {
  int conf_fd = sceIoOpen( conf_name, PSP_O_RDONLY, 0644 );
  if ( conf_fd >= 0 ) {
    int bytes=0;int line; char buf[256];
    for( line = 0; line < CFG_MAX_ENTRIES; line++ ) {
      bytes=readLine(conf_fd, buf, 256);
      if ( (bytes > 5) && (cfg_count <= CFG_MAX_ENTRIES ) ) {
        strcpy(buf,strstrip(buf));
        if (strncmp(buf, "#", 1)) {
          log2("cfg line %d: '%s'\n",line,buf);
          char *modname;char *bootname;
          modname = strtok(buf, " \t\n\r");
          bootname = strtok(0, "#\r\n");
          if(bootname) {
            char tmodname[strlen(modname)+5];
            if(strncmp(modname, "/",1)==0) {
              strcpy(tmodname,"ms0:");
              strcat(tmodname,modname);
            } else {
              strcpy(tmodname,modname);
            }
            strcpy(bootname,strstrip(bootname));
            char tbootname[strlen(bootname)+5];
            if(strncmp(bootname,"/",1)==0) {
              strcpy(tbootname,"ms0:");
              strcat(tbootname,bootname);
            } else if ( (strlen(bootname) == 10) && ( bootname[4] == '-' ) ) {
              strncpy(tbootname,bootname,4);
              strncpy(tbootname+4,bootname+5,6);
            } else {
              strcpy(tbootname,bootname);
            }
            strncpy(cfg[cfg_count].modname, tmodname,strlen(tmodname));
            strncpy(cfg[cfg_count].bootname, tbootname,strlen(tbootname));
            cfg_count++;
            if ( cfg_count % 32 == 0 ) {
              Cfg * tmp = sceKernelAllocHeapMemory( cfg_mem_id, sizeof( Cfg ) * ( cfg_count + 32 ) );
              memcpy( tmp, cfg, sizeof( Cfg ) * cfg_count );
              sceKernelFreeHeapMemory( cfg_mem_id, cfg );
              cfg = tmp;
            }
          } else {
            if (strncmp(modname,"logging=1",9)==0) {
              logtofile=1;
              log("Logging activated.\n");
            } else if (strncmp(modname,"logging=2",9)==0) {
              logtofile=2;
              log("Verbose logging activated.\n");
            } else {
              log("Syntax error on line %d: ('%s', '%s')\n",line,modname,bootname);
            }
          }
        }
      }
      if ( bytes < 0 ) { 
        break;
      }
    }
    sceIoClose(conf_fd);    
  }
} 


///////////////////////////////////
///// Load a module and blacklist it
int loadmod(const char *name, int argc, char **argv) {
  int i;
  for ( i = 0; i < loaded_count; i ++ ) {
    if ( strncmp( name, loaded[i].modname, strlen(loaded[i].modname) ) == 0 ) {
      log("  module %s already loaded (or attempted to).\n",name);
      return -1;
    }
  }

  strncpy(loaded[loaded_count].modname, name,strlen(name));
  loaded_count++;

  SceUID modid;
  int status;
  char args[1024];
  modid = sceKernelLoadModule(name, 0, NULL);
  if(modid >= 0) {
    modid = sceKernelStartModule(modid, (strlen(name) + 1), (void *) args, &status, NULL);
    if (logtofile < 2) log("  start module %s: %08x\n", name, modid );
  } else {
    log("  load module %s - error: %08x\n", name, modid );
  }
  return modid;
}
///////////////////////////////////
///// Have sceMediaSync poke main thread. FIXME: Can't turn this off?
int OnModuleStart(SceModule2* mod) {
  log2("Start - %s\n", mod->modname );
  //if ( strcmp( mod->modname, "sceKernelLibrary" ) == 0 ) { FIXME - bail somehow }
  if ( strcmp( mod->modname, "sceMediaSync" ) == 0 ) {
    sceKernelSignalSema( sema, 1 );
  }
  return previous((!previous)?0:mod);
}


///////////////////////////////////
///// old method of getting umdID via disc0:/UMD_DATA.BIN...doesn't work for real UMDs
int getumdinfo( char * buf ) {
  int fdUMD = sceIoOpen("disc0:/UMD_DATA.BIN",PSP_O_RDONLY,0);
  if (fdUMD >= 0) {
    sceIoRead(fdUMD,buf,10);
    buf[10] = 0;
    sceIoClose(fdUMD);
    log("  disc0:/UMD_DATA.BIN - id: %s.\n",buf);
    return 1;
  } else {
    log("  disc0:/UMD_DATA.BIN - cannot open (0x%08X).\n",fdUMD);
    return 0;
  }
}

///////////////////////////////////
///// case insensitive memory read
int readmemchar ( char * buf, unsigned int addr , unsigned int len ) {
  int c;char byte;
  strcpy(buf,"_");
  for ( c=0; c < len ; c++ ) {
    byte=peekbyte(addr+c);
    if (byte == '\0') {
      buf[c]=byte;
      break;
    } else {
      buf[c]=(char)tolower(byte);
      buf[c+1]='\0';
    }
  }
  log("readmem: Fetched '%s' from 0x%08X, %d bytes.\n",buf,addr,c);
  return len;
}
///////////////////////////////////
///// yoink iso filename from SystemControl
int isopath ( char * buf ) {
  SceModule * sysctrl = ( SceModule * )sceKernelFindModuleByName( ISOMODULE );
  if (sysctrl) {
    SceKernelModuleInfo sysctrlmodinfo;
    unsigned int addr, isoaddr, offset;
    sysctrlmodinfo.size=sizeof(SceKernelModuleInfo);
    sceKernelQueryModuleInfo(sysctrl->modid, &sysctrlmodinfo);
    addr=sysctrlmodinfo.text_addr;
    offset=ISOOFFSET;
    isoaddr=addr + offset;
    log2("isopath: Found (0x%08X) %s, text_addr: 0x%08X + offset: 0x%08X = isoaddr: 0x%08X\n",sysctrl->modid, sysctrl->modname, addr, offset, isoaddr);
    readmemchar(buf,isoaddr,64);
    if (strlen(buf) < 5) return 0;
    return 1;
  }
  return 0;
}

///////////////////////////////////
///// need better method of getting mitts into sceMediaSync
int install( void ) {
  log2("install: sctrlHENSetStartModuleHandler...\n");
  previous = sctrlHENSetStartModuleHandler(OnModuleStart);
  return 0;
}

///////////////////////////////////
///// load appropriate modules for 'oid' from cfg struct
int loadmods( const char * oid ) {
  //log2("loadmods - %s\n",oid);
  if (strlen(oid) < 3) return 0;

  char id[strlen(oid)];
  strcpy(id,oid);
  int si, i;

  for (si = 0; id[si] != '\0'; si++) {
    id[si] = (char)tolower(id[si]);
  }

  for ( i = 0; i < cfg_count; i ++ ) {
    if (strncmp( id, cfg[i].bootname, strlen(cfg[i].bootname) ) == 0 ) {
      log("match: %s <=> %s: %s\n",id,cfg[i].bootname,cfg[i].modname);
      loadmod(cfg[i].modname, 0, NULL);
    }
  }
  return 0;
}
//////////////////////////////
///// load modules from PATH.EXT.plugins or PATH.plugins

int loadlocalmods(const char * opath) {
  char path[strlen(opath)+8];
  strcpy(path,opath);

  strcat(path,".plugins");
  log2("Loadlocal - trying %s / %s\n",opath,path);
  
  int lconf_fd = sceIoOpen( path, PSP_O_RDONLY, 0644 );
  if ( (lconf_fd < 0 ) && (path[strlen(opath)-4] == '.' ) ) {
    strncpy(path+(strlen(opath)-4),".plugins",8);
    path[strlen(opath)+4]= '\0';
    log2("Loadlocal - trying %s / %s\n",opath,path);
    lconf_fd = sceIoOpen( path, PSP_O_RDONLY, 0644 );
  }
  if (lconf_fd >= 0 ) {
    log("Loadlocal - found %s.\n",path);
    int bytes=0;int line; char buf[256];
    for( line = 0; line < CFG_MAX_ENTRIES; line++ ) {
      bytes=readLine(lconf_fd, buf, 256);
      if (bytes > 5) {
        strcpy(buf,strstrip(buf));
        if (strncmp(buf, "#", 1)) {
          char tmodname[strlen(buf)+5];
          if(strncmp(buf, "/",1)==0) {
            strcpy(tmodname,"ms0:");
            strcat(tmodname,buf);
          } else {
            strcpy(tmodname,buf);
          }
          log("%s: loading '%s'.\n",path,tmodname);
          loadmod(buf,0,NULL);
        }
      }
    }
    sceIoClose(lconf_fd);
  }

  return 0;
}
//////////////////////////////
///// sleep until sceMediaSync starts, then load stuff.
int main_thread ( SceSize args, void *argp ) {
  sema = sceKernelCreateSema( "pergame_wake", 0, 0, 1, NULL );
  sceKernelWaitSemaCB( sema, 1, NULL );
  sceKernelDeleteSema(sema); sema = -1;
  log("main: Thread wake. %d cfg entries.\n",cfg_count);
  if(logtofile >= 2) {
    int i;
    for ( i = 0; i < cfg_count; i ++ ) {
      log2("cfg[%d]: modname: %s, bootname: %s\n", i, cfg[i].modname, cfg[i].bootname);    
    }
  }
  										//char b[12];readmem(b,0x88013780+48,11);// works with UMD or ISO
  if (sceKernelBootFrom() == PSP_BOOT_DISC) {
    int isoread=0; int umdread=0;
    char iso[64];
    char umdID[11];
    isoread=isopath(iso);
#ifdef USE_UMD_DATA_BIN
    umdread=getumdinfo(umdID);
#else
    umdread=readmemchar(umdID,UMDIDADDR,11);
#endif
    log("main: bootfrom: %04X, apitype: %04X, keyconf: %04X, initfile: %s, iso: '%s', umdid: '%s'\n", bootdev, apitype, keyconf, eboot,
      (isoread?iso:"-error-"),(umdread?umdID:"-error-"));
    if(isoread) {
      loadlocalmods(iso);
      loadmods(iso);
      loadmods("iso");
    } else {
      loadmods("umd");
    }
    if(umdread) loadmods(umdID);
  } else {
    log("main: bootfrom: %04X, apitype: %04X, keyconf: %04X, initfile: '%s'\n",bootdev, apitype, keyconf, eboot);
    loadlocalmods(eboot);
    loadmods(eboot);
  }
  log2("main: cleanup start.\n");
  sceKernelFreeHeapMemory( cfg_mem_id, cfg );
  sceKernelFreeHeapMemory( cfg_mem_id, loaded );
  sceKernelDeleteHeap( cfg_mem_id );
  cfg = NULL;
  loaded = NULL;
  log("main: cleanup done.\n");

//  sceKernelDelayThread(100000);
//  sceKernelExitDeleteThread(0);
  
  return 0;
}

///////////////////////////////////
///// main()
int module_start(SceSize args, char *argp) {
  //logtofile=1;
  char * cwdp;
  strcpy(conf_name,argp);
  cwdp = strrchr(conf_name, '/');
  *cwdp = '\0';
  strcpy(log_name,conf_name);
  strcat(conf_name,"/pergame.txt");
  strcat(log_name,"/pergame.log");
  sceIoRemove(log_name);
  log("\n==========\nstart. conf: %s, free kmem: %d\n",conf_name,sceKernelPartitionTotalFreeMemSize(PSP_MEMORY_PARTITION_KERNEL));

  cfg_mem_id = sceKernelCreateHeap( PSP_MEMORY_PARTITION_KERNEL, 
      ( sizeof( Cfg ) * ( CFG_MAX_ENTRIES + 1) ) + 
      ( sizeof( Loaded ) * 17 )
    , 1, "cfg_heap");
  cfg = sceKernelAllocHeapMemory( cfg_mem_id, sizeof( Cfg ) * 32 );
  memset( cfg, 0, sizeof( Cfg ) * 32 );

  loaded = sceKernelAllocHeapMemory( cfg_mem_id, sizeof( Loaded ) * 16 );
  memset( cfg, 0, sizeof( Loaded ) * 16 );

  parse_conf();

  if (cfg_count <= 0) return 0; // no point

  eboot = sceKernelInitFileName(); 
  bootdev = sceKernelBootFrom(); // PSP_BOOT_FLASH = 0, PSP_BOOT_DISC = 0x20, PSP_BOOT_MS = 0x40
  apitype = sceKernelInitApitype(); // PSP_INIT_KEYCONFIG_VSH = 0x100, PSP_INIT_KEYCONFIG_GAME = 0x200, PSP_INIT_KEYCONFIG_POPS = 0x300
  keyconf = InitForKernel_7233B5BC();

  log2("module_start - bootfrom: %04X, apitype: %04X, keyconf: %04X, initfile: %s\n", bootdev, apitype, keyconf, eboot);

										  //char b[12];readmemchar(b,0x880137B0,11);//0x880137B0

  install();

  int th = sceKernelCreateThread("pergame", main_thread, 0x18, 128*150, 0, NULL);// 8, 64*1024, 0, NULL);// 47, 0x400, 0x00001001, NULL);
  if( th >= 0 )
    sceKernelStartThread(th, 0, 0);
  return 0;
}
///////////////////////////////////


int module_stop(SceSize args, void *argp) {
  log("Baibai!\n");
  return 0;
}
